<?php

namespace CloudPrinterPlugin\Controllers;

use CloudPrinter\CloudApps\Client\CloudAppsClient;
use CloudPrinterPlugin\Helpers\CloudPrinterHelper;
use CloudPrinterPlugin\Libs\BaseController;

/**
 * Class ReorderController
 *
 * @copyright   2019 by CloudPrinter
 * @author      Vasyl Kushniruk <vk@cloudprinter.com>
 */
class ReorderController extends BaseController
{
    public function __construct()
    {
        parent::__construct(CP_ABS_PATH, 'reorder');

        add_action('rest_api_init', function () {
            register_rest_route('cloudprinter/v1', '/reorder', [
                'methods' => 'POST',
                'permission_callback' => [$this, 'verifyApiRequest'],
                'callback' => [$this, 'makeReorder']
            ]);
        });

        add_action('rest_api_init', function () {
            register_rest_route('cloudprinter/v1', '/cp-order/shipping', [
                'methods' => 'POST',
                'permission_callback' => [$this, 'verifyApiRequest'],
                'callback' => [$this, 'getShipping']
            ]);
        });
    }

    public function verifyApiRequest()
    {
        $headers = array_change_key_case(getallheaders());

        return wp_verify_nonce($headers['x-wp-nonce'], 'wp_rest') && current_user_can( 'edit_posts' );
    }

    public function makeReorder($requestData)
    {
        $originalOrderId = $requestData['order_id'];
        $itemKeys = $requestData['reorder_item_references'];

        if (!$originalOrderId) {
            return ['error' => 'Something went wrong'];
        }

        // Check if some product selected
        if (!$itemKeys) {
            return ['error' => 'No products selected'];
        }

        $originalOrder = new \WC_Order($originalOrderId);
        $currentUser = wp_get_current_user();

        // Check if user has permission
        if (!current_user_can('edit_posts')) {
            return ['error' => 'Permission denied'];
        };

        if (!$this->orderIsReordable($originalOrderId, $itemKeys)) {
            return ['error' => 'The reorder is not possible'];
        }

        // Create Order
        $orderData = array(
            'post_type' => 'shop_order',
            'post_title' => sprintf(__('Reorder &ndash; %s', 'woocommerce'), strftime(_x('%b %d, %Y @ %I:%M %p', 'Order date parsed by strftime', 'woocommerce'))),
            'post_status' => 'draft',
            'ping_status' => 'closed',
            'post_excerpt' => 'Reorder based on original order ' . $originalOrderId,
            'post_author' => $currentUser->ID,
            'post_password' => uniqid('order_')   // Protects the post just in case
        );
        $orderId = wp_insert_post($orderData, true);

        if (is_wp_error($orderId)) {
            $errorMessage = "Unable to create order:" . $orderId->get_error_message();
            throw new \Exception($errorMessage);
        } else {
            $order = wc_get_order($orderId);

            // Copy custom metadata
            $data = get_post_custom($originalOrderId);
            foreach ($data as $key => $values) {
                foreach ($values as $value) {
                    add_post_meta($orderId, $key, $value);
                }
            }

            if ($requestData['reorder-type'] == 'reorder-certain-items') {
                if (empty($requestData['shipping-methods'])) {
                    return ['error' => 'Shipping is not selected'];
                }

                $shippingMethods = [];
                foreach ($requestData['shipping-methods'] as $key => $shippingMethod) {
                    $shippingMethods[$key] = json_decode($shippingMethod, true);
                }

                $order->update_meta_data('_cp_shipping_data', json_encode($shippingMethods));
            } else {
                $order->update_meta_data('_cp_shipping_data', $originalOrder->get_meta('_cp_shipping_data', true));
            }

            // Update order meta data
            $order->update_meta_data('_reorder_origin_order_id', $originalOrderId);
            $order->update_meta_data('_cp_items_canceled', '');
            $order->update_meta_data('_shipping_refund_amount', null);
            $order->update_meta_data('_cp_status', '');
            $order->update_meta_data('_cp_order_json', '');
            $order->update_meta_data('_cp_order_quote', '');
            $order->update_meta_data('_cp_tracking_info', '');
            $order->update_meta_data('_cp_reorder', 1);
            $order->update_meta_data('_order_shipping', $originalOrder->get_meta('_order_shipping', true));
            $order->update_meta_data('_order_discount', $originalOrder->get_meta('_order_discount', true));
            $order->update_meta_data('_cart_discount', $originalOrder->get_meta('_cart_discount', true));
            $order->update_meta_data('_order_tax', $originalOrder->get_meta('_order_tax', true));
            $order->update_meta_data('_order_shipping_tax', $originalOrder->get_meta('_order_shipping_tax', true));
            $order->update_meta_data('_order_total', 0);
            $order->update_meta_data('_order_key', 'wc_' . apply_filters('woocommerce_generate_order_key', uniqid('order_')));
            $order->update_meta_data('_customer_user', $originalOrder->get_meta('_customer_user', true));
            $order->update_meta_data('_order_currency', $originalOrder->get_meta('_order_currency', true));
            $order->update_meta_data('_prices_include_tax', $originalOrder->get_meta('_prices_include_tax', true));
            $order->update_meta_data('_customer_ip_address', $originalOrder->get_meta('_customer_ip_address', true));

            // Add billing fields
            $order->update_meta_data('_billing_city', $originalOrder->get_meta('_billing_city', true));
            $order->update_meta_data('_billing_state', $originalOrder->get_meta('_billing_state', true));
            $order->update_meta_data('_billing_postcode', $originalOrder->get_meta('_billing_postcode', true));
            $order->update_meta_data('_billing_email', $originalOrder->get_meta('_billing_email', true));
            $order->update_meta_data('_billing_phone', $originalOrder->get_meta('_billing_phone', true));
            $order->update_meta_data('_billing_address_1', $originalOrder->get_meta('_billing_address_1', true));
            $order->update_meta_data('_billing_address_2', $originalOrder->get_meta('_billing_address_2', true));
            $order->update_meta_data('_billing_country', $originalOrder->get_meta('_billing_country', true));
            $order->update_meta_data('_billing_first_name', $originalOrder->get_meta('_billing_first_name', true));
            $order->update_meta_data('_billing_last_name', $originalOrder->get_meta('_billing_last_name', true));
            $order->update_meta_data('_billing_company', $originalOrder->get_meta('_billing_company', true));

            // Add Shipping Fields
            $order->update_meta_data('_shipping_country', $originalOrder->get_meta('_shipping_country', true));
            $order->update_meta_data('_shipping_first_name', $originalOrder->get_meta('_shipping_first_name', true));
            $order->update_meta_data('_shipping_last_name', $originalOrder->get_meta('_shipping_last_name', true));
            $order->update_meta_data('_shipping_company', $originalOrder->get_meta('_shipping_company', true));
            $order->update_meta_data('_shipping_address_1', $originalOrder->get_meta('_shipping_address_1', true));
            $order->update_meta_data('_shipping_address_2', $originalOrder->get_meta('_shipping_address_2', true));
            $order->update_meta_data('_shipping_city', $originalOrder->get_meta('_shipping_city', true));
            $order->update_meta_data('_shipping_state', $originalOrder->get_meta('_shipping_state', true));
            $order->update_meta_data('_shipping_postcode', $originalOrder->get_meta('_shipping_postcode', true));

            // For order note.
            $reorderDetails = [];

            // Add line items
            foreach ($originalOrder->get_items() as $originItemId=> $originalOrderItem) {
                if (!in_array($originalOrderItem['cartItemKey'], $itemKeys)) {
                    continue;
                }

                if (!array_key_exists($originalOrderItem['cartItemKey'], $requestData['reorder_quantity'])) {
                    $errorMessage = "Unable to make reorder.";
                    return ['error' => $errorMessage];
                }

                $itemName = $originalOrderItem['name'];
                $lineTotal = 0;
                $productId = $originalOrderItem['product_id'];
                $qty = $requestData['reorder_quantity'][$originalOrderItem['cartItemKey']];
                $cartItemKey = wc_get_order_item_meta($originItemId, '_cartItemKey');

                $itemId = wc_add_order_item($orderId, [
                    'order_item_name' => $itemName,
                    'order_item_type' => 'line_item'
                ]);

                wc_add_order_item_meta($itemId, '_qty', $qty);
                wc_add_order_item_meta($itemId, '_product_id', $productId);
                wc_add_order_item_meta($itemId, '_line_subtotal', wc_format_decimal($lineTotal));
                wc_add_order_item_meta($itemId, '_line_total', wc_format_decimal($lineTotal));
                wc_add_order_item_meta($itemId, '_line_tax', wc_format_decimal('0'));
                wc_add_order_item_meta($itemId, '_line_subtotal_tax', wc_format_decimal('0'));
                wc_add_order_item_meta($itemId, '_projectFileName', wc_get_order_item_meta($originItemId, '_projectFileName'));
                wc_add_order_item_meta($itemId, '_cartItemKey', $cartItemKey);
                wc_add_order_item_meta($itemId, '_dynamicOptions', wc_get_order_item_meta($originItemId, '_dynamicOptions'));

                wc_add_order_item_meta($itemId, '_reorder_cause', $requestData['reorder_cause'][$cartItemKey]);
                wc_add_order_item_meta($itemId, '_reorder_description', $requestData['reorder_description'][$cartItemKey]);
                wc_add_order_item_meta($originItemId, '_reorder_cause', $requestData['reorder_cause'][$cartItemKey]);
                wc_add_order_item_meta($originItemId, '_reorder_description', $requestData['reorder_description'][$cartItemKey]);

                $reorderDetails[] =[
                    'product_name' => $itemName,
                    'cause' => $requestData['reorder_cause'][$cartItemKey],
                    'description' => $requestData['reorder_description'][$cartItemKey]
                ];
            }

            // Store coupons
            $originalOrderCoupons = $originalOrder->get_items('coupon');
            foreach ($originalOrderCoupons as $originalOrderCoupon) {
                $itemId = wc_add_order_item($orderId, array(
                    'order_item_name' => $originalOrderCoupon['name'],
                    'order_item_type' => 'coupon'
                ));
                // Add line item meta
                if ($itemId) {
                    wc_add_order_item_meta($itemId, 'discount_amount', $originalOrderCoupon['discount_amount']);
                }
            }

            // Payment info
            $order->update_meta_data('_payment_method', $originalOrder->get_meta('_payment_method', true));
            $order->update_meta_data('_payment_method_title', $originalOrder->get_meta('_payment_method_title', true));
            $order->update_meta_data('Transaction ID', $originalOrder->get_meta('Transaction ID', true));
            $order->payment_complete();

            // Set order status to processing to trigger initial emails to end user and vendor
            $updateNote = $this->getReorderNote($originalOrderId, $reorderDetails);
            $order->update_status('processing');
            $order->add_order_note($updateNote);

            $order->set_total(0);
            $order->save();
        }
        return ['order_id' => $orderId];
    }

    private function getReorderNote($originalOrderId, $reorderDetails)
    {
        $result = sprintf('Order has been created from reordering. Original order: %1$s. ', $originalOrderId);

        foreach ($reorderDetails as $reorderDetail) {
            $result .= sprintf(
                'Product name: %1$s, reorder cause: "%2$s", reorder description: "%3$s"',
                $reorderDetail['product_name'],
                $reorderDetail['cause'],
                $reorderDetail['description']
            );
        }

        return $result;
    }

    public function getShipping($requestData)
    {
        $orderId = $requestData['order_id'];

        if (!$orderId) {
            return new \WP_REST_Response(['error' => 'Order id is not set.'], 422);
        }

        $order = new \WC_Order($orderId);
        try {
            $orderQuote = CloudPrinterHelper::requestOrderQuotes($order, $requestData['reorder_item_references'], $requestData['reorder_quantity']);
        } catch (\Exception $e) {
            return ['error' => 'Something went wrong'];
        }

        $result = [
            'html' => $this->getHtml('shipping-method', [
                'orderQuote' => $orderQuote,
                'order' => $order
            ])
        ];

        return $result;
    }

    private function orderIsReordable($orderId, $orderItems = [])
    {
        $order = wc_get_order($orderId);
        $orderReference = $order->get_meta('_cp_order_reference', true);
        $accessToken = get_option('cp_api_key');
        $cloudPrinterClient = new CloudAppsClient($accessToken);

        $orderInfoResponse = $cloudPrinterClient->order->getInfo($orderReference);
        $orderInfoData = $orderInfoResponse->getData();

        foreach ($orderInfoData['items'] as $item) {
            if (!empty($orderItems) && !in_array($item['reference'], $orderItems)) {
                continue;
            }

            if (!$item['is_reordable']) {
                return false;
            }
        }

        return  true;
    }
}
